@@ -2,10 +2,11 @@
 
 from __future__ import absolute_import, division, print_function
 
-import os
-
 import pytest
 
+import datetime
+import os
+
 from pywaterinfo.waterinfo import Waterinfo
 
 # use a session for VMM that can be used among tests
@@ -17,6 +18,12 @@ vmm_client = (
 hic_client = os.environ.get("HIC_TOKEN")
 
 
+@pytest.fixture
+def patch_retention(monkeypatch):
+    retention = datetime.timedelta(seconds=1)
+    monkeypatch.setattr("pywaterinfo.waterinfo.CACHE_RETENTION", retention)
+
+
 @pytest.fixture(scope="module")
 def vmm_connection():
     return Waterinfo("vmm", token=vmm_client)
